/* 
  Copyright 2004 Paul Zimmermann.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/resource.h>

#include "gmp.h"
#include "mpfr.h"
#include "wclr.h"

#ifdef WITH_MPI
#include "mpi.h"
#endif

#define VERSION "1.6.1"

#define OUT(x) mpz_out_str (stdout, 10, x); printf ("\n")

#define ROWS 6
#define EXTRA 2 /* maximal allowed width is 2^EXTRA*T */

#define Rt0u0 1 /* 1 */
#define Rt1u0 2 /* T*t */
#define Rt2u0 3 /* (T*t)^2 */
#define Rt3u0 4 /* (T*t)^3 */
#define Rt4u0 5 /* (T*t)^4 */
#define Rt0u1 6 /* P'+3u */
#define Rt1u1 7 /* (T*t)*(P'+3u) */
#define Rt2u1 8 /* (T*t)^2*(P'+3u) */
#define Rt0u2 9 /* (P'+3u)^2 */

/* #define VERBOSE */

void wclr_mpz_set_fr (mpz_t, mpfr_t);
mpz_t* init_vec (unsigned long);
void clear_vec (mpz_t *, unsigned long);
void set_vec (mpz_t *, mpz_t *, unsigned long);
void init_matrix (mat_ZZ *, unsigned long);
void clear_matrix (mat_ZZ *, unsigned long);
void Mod (mpz_t *, int, mpz_t);
void MakeMonic (mpz_t *, int, mpz_t);
int Reduce (mpz_t *, unsigned long, mpz_t *, int, mpz_t);
unsigned long GCD (mpz_t *, int, mpz_t *, int, mpz_t);
int Powmod (mpz_t *, int, mpz_t, mpz_t *, int, mpz_t);
unsigned long DDF (mpz_t *, unsigned long, mpz_t);
void centermod (mpz_t *, int, mpz_t);
void Split (mpz_t *, int, mpz_t);
unsigned int Roots (mpz_t *, int, mpz_t);
int Check (mpz_t *, int, mpz_t, unsigned long int, long int);
void lift_matrix (mpz_t **, unsigned int, mpz_t);

#define MALLOC malloc
#define FREE free

static int
cputime ()
{
  struct rusage rus;

  getrusage (0, &rus);
  return rus.ru_utime.tv_sec * 1000 + rus.ru_utime.tv_usec / 1000;
}

int lll_time;

/* return the degree of r[0]+t*r[1]+...+t^n*r[n], -1 if identically 0 */
static int
Degree (mpz_t *r, int n)
{
  while (n >= 0 && mpz_cmp_ui (r[n], 0) == 0)
    n--;
  return n;
}

#if 0
/* put in r[0..4] the resultant of a and b (mod q) [multiplied by c^4],
   where a = a1+a2*t+a3*t^2+a4*t^3+a5*t^4+a6*u+a7*u*t+a8*u*t^2+a9*u^2
           where a7 = 3*a4/c, a8 = 6*a5/c, a9 = 9*a5/c^2.
         b = b1+b2*t+b3*t^2+b4*t^3+b5*t^4+b6*u+b7*u*t+b8*u*t^2+b9*u^2
           where b7 = 3*b4/c, b8 = 6*b5/c, b9 = 9*b5/c^2.

The coefficients were computed with Maple as follows:
a[7]:=3*a[4]/c: a[8]:=6*a[5]/c: a[9]:=9*a[5]/c^2:
b[7]:=3*b[4]/c: b[8]:=6*b[5]/c: b[9]:=9*b[5]/c^2:
A:=a[1]+a[2]*t+a[3]*t^2+a[4]*t^3+a[5]*t^4+a[6]*u+a[7]*u*t+a[8]*u*t^2+a[9]*u^2:
B:=b[1]+b[2]*t+b[3]*t^2+b[4]*t^3+b[5]*t^4+b[6]*u+b[7]*u*t+b[8]*u*t^2+b[9]*u^2:
r:=resultant(A,B,u):
r:=normal(c^4*r/9):
r2:=collect(coeff(r,c,2),t,factor):
r1:=collect(coeff(r,c,1),t,factor):
r0:=collect(coeff(r,c,0),t,factor):
l:=[codegen[optimize]([seq(R[i]=coeff(r0+r1+r2,t,i),i=0..4)])]:
# suppress affectations ti=a[j]
l1 := select(type, l, name=indexed):
l2 := remove(type, l, name=indexed):
l := subs(l1, l2):
read `gmp.mpl`;
for i to 1 do gmp_set(op(l[i])) od:
*/
static void
Res22new (mpz_t *r, mpz_t *a, mpz_t *b, mpz_t q, mpz_t tmp, mpz_t c)
{
#define Ntmp 42
  mpz_t t[Ntmp];
  int i;

  /* first compute a[7..9] */
  mpz_divexact (a[7], a[4], c);  /* a4/c */
  mpz_mul_ui (a[7], a[7], 3);    /* 3*a4/c */
  mpz_divexact (a[8], a[5], c);  /* a5/c */
  mpz_divexact (a[9], a[8], c);  /* a5/c^2 */
  mpz_mul_ui (a[8], a[8], 6);    /* 6*a5/c */
  mpz_mul_ui (a[9], a[9], 9);    /* 9*a5/c^2 */

  /* then compute b[7..9] */
  mpz_divexact (b[7], b[4], c);  /* b4/c */
  mpz_mul_ui (b[7], b[7], 3);    /* 3*b4/c */
  mpz_divexact (b[8], b[5], c);  /* b5/c */
  mpz_divexact (b[9], b[8], c);  /* b5/c^2 */
  mpz_mul_ui (b[8], b[8], 6);    /* 6*b5/c */
  mpz_mul_ui (b[9], b[9], 9);    /* 9*b5/c^2 */

  for (i=0;i<Ntmp;i++)
    mpz_init (t[i]);
#define t200 t[0]
#define t3 t[1]
#define t6 t[2]
#define t7 t[3]
#define t8 t[4]
#define t11 t[5]
#define t15 t[6]
#define t16 t[7]
#define t17 t[8]
#define t20 t[9]
#define t22 t[10]
#define t26 t[11]
#define t28 t[12]
#define t31 t[13]
#define t32 t[14]
#define t35 t[15]
#define t42 t[16]
#define t45 t[17]
#define t49 t[18]
#define t53 t[19]
#define t54 t[20]
#define t57 t[21]
#define t58 t[22]
#define t62 t[23]
#define t71 t[24]
#define t76 t[25]
#define t79 t[26]
#define t82 t[27]
#define t83 t[28]
#define t86 t[29]
#define t89 t[30]
#define t90 t[31]
#define t93 t[32]
#define t96 t[33]
#define t111 t[34]
#define t114 t[35]
#define t117 t[36]
#define t121 t[37]
#define t140 t[38]
#define t143 t[39]
#define t174 t[40]
#define t177 t[41]
   mpz_mul (t3, b[1], a[5]);
   mpz_mul (t6, b[5], a[1]);
   mpz_sub (t7, t3, t6);
   mpz_mul (t8, t7, t7);
   mpz_mul (t11, b[6], a[1]);
   mpz_mul (t15, b[5], a[6]);
   mpz_mul (t16, a[5], b[6]);
   mpz_sub (t17, t16, t15);
   /* r[0] = 9*t8+(t11-a[6]*b[1])*t17 */
   mpz_mul_ui (r[0], t8, 9);
   mpz_mul (tmp, a[6], b[1]);
   mpz_sub (tmp, t11, tmp);
   mpz_addmul (r[0], tmp, t17);
   mpz_mod (r[0], r[0], q);

   mpz_mul (t20, a[2], b[5]);
   mpz_mul (t22, a[5], b[2]);
   mpz_mul (t26, b[6], b[5]);
   mpz_mul (t28, a[4], a[1]);
   mpz_mul (t31, b[1], b[5]);
   mpz_mul (t32, a[6], a[4]);
   mpz_mul (t35, a[4], b[6]);
   mpz_mul (t42, a[6], b[4]);
   mpz_mul (t45, b[4], b[5]);
   mpz_mul (t49, b[6], a[2]);
   /* r[1] = 18*(-t20+t22)*t7-3*t26*t28+6*t31*t32-3*t35*t3+6*t16*b[4]*a[1]
      -3*t42*t3-3*t45*a[6]*a[1]+(t49-a[6]*b[2])*t17 */
   mpz_sub (tmp, t22, t20);
   mpz_mul (tmp, tmp, t7);
   mpz_mul_si (r[1], tmp, -18);
   mpz_mul (tmp, t26, t28);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, t31, t32);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[1], r[1], tmp);
   mpz_mul (tmp, t35, t3);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, t16, b[4]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[1], r[1], tmp);
   mpz_mul (tmp, t42, t3);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, t45, a[6]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, a[6], b[2]);
   mpz_sub (tmp, t49, tmp);
   mpz_addmul (r[1], tmp, t17);
   mpz_mod (r[1], r[1], q);

   mpz_mul (t53, b[4], b[4]);
   mpz_mul (t54, t53, a[5]);
   mpz_mul (t57, b[5], b[5]);
   mpz_mul (t58, a[2], a[2]);
   mpz_mul (t62, a[5], b[3]);
   mpz_mul (t71, b[4], a[4]);
   mpz_mul (t76, a[4], a[4]);
   mpz_mul (t79, b[5], a[3]);
   mpz_mul (t82, a[5], a[5]);
   mpz_mul (t83, b[2], b[2]);
   mpz_mul (t86, t82, b[3]);
   /* t89 = 9*t54*a[1]+9*t57*t58-18*t6*t62+18*t57*a[1]*a[3]-18*t20*t22
      -9*t71*t3-9*t45*t28+9*t31*t76-18*t79*t3+9*t82*t83+18*t86*b[1] */
   mpz_mul (tmp, t54, a[1]);
   mpz_mul_ui (t89, tmp, 9);
   mpz_mul (tmp, t57, t58);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (t89, t89, tmp);
   mpz_mul (tmp, t6, t62);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_sub (t89, t89, tmp);
   mpz_mul (tmp, t57, a[1]);
   mpz_mul (tmp, tmp, a[3]);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_add (t89, t89, tmp);
   mpz_mul (tmp, t20, t22);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_sub (t89, t89, tmp);
   mpz_mul (tmp, t71, t3);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_sub (t89, t89, tmp);
   mpz_mul (tmp, t45, t28);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_sub (t89, t89, tmp);
   mpz_mul (tmp, t31, t76);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (t89, t89, tmp);
   mpz_mul (tmp, t79, t3);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_sub (t89, t89, tmp);
   mpz_mul (tmp, t82, t83);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (t89, t89, tmp);
   mpz_mul (tmp, t86, b[1]);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_add (t89, t89, tmp);
   mpz_mul (t90, b[5], a[5]);
   mpz_mul (t93, a[4], a[2]);
   mpz_mul (t96, b[2], b[5]);
   mpz_mul (t111, t57, a[6]);
   mpz_mul (t114, t82, b[6]);
   mpz_mul (t117, b[6], a[3]);
   /* t121 = 6*t90*t11-3*t26*t93+6*t96*t32-3*t45*a[6]*a[2]-3*t35*t22+6*t15*t3
      +6*t16*b[4]*a[2]-3*t42*t22-6*t111*a[1]-6*t114*b[1]+(t117-a[6]*b[3])*t17*/
   mpz_mul (tmp, t90, t11);
   mpz_mul_ui (t121, tmp, 6);
   mpz_mul (tmp, t26, t93);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (t121, t121, tmp);
   mpz_mul (tmp, t96, t32);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (t121, t121, tmp);
   mpz_mul (tmp, t45, a[6]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (t121, t121, tmp);
   mpz_mul (tmp, t35, t22);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (t121, t121, tmp);
   mpz_mul (tmp, t15, t3);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (t121, t121, tmp);
   mpz_mul (tmp, t16, b[4]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (t121, t121, tmp);
   mpz_mul (tmp, t42, t22);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (t121, t121, tmp);
   mpz_mul (tmp, t111, a[1]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_sub (t121, t121, tmp);
   mpz_mul (tmp, t114, b[1]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_sub (t121, t121, tmp);
   mpz_mul (tmp, a[6], b[3]);
   mpz_sub (tmp, t117, tmp);
   mpz_addmul (t121, tmp, t17);
   mpz_add (r[2], t89, t121);
   mpz_mod (r[2], r[2], q);

   mpz_mul (t140, b[3], b[5]);
   mpz_mul (t143, a[4], a[3]);
   /* r[3] = -9*t71*t22+18*t57*a[3]*a[2]+18*t86*b[2]-18*t79*t22
      -18*t90*b[3]*a[2]+9*t96*t76-9*t45*t93+9*t54*a[2]+6*t140*t32-3*t26*t143
      +6*t16*b[4]*a[3]-3*t35*t62-3*t42*t62-6*t114*b[2]-3*t45*a[6]*a[3]
      +6*t90*t49-6*t111*a[2]+6*t15*t22+t17*(t35-t42) */
   mpz_mul (tmp, t71, t22);
   mpz_mul_si (r[3], tmp, -9);
   mpz_mul (tmp, t57, a[3]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t86, b[2]);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t79, t22);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t90, b[3]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t96, t76);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t45, t93);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t54, a[2]);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t140, t32);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t26, t143);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t16, b[4]);
   mpz_mul (tmp, tmp, a[3]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t35, t62);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t42, t62);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t114, b[2]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t45, a[6]);
   mpz_mul (tmp, tmp, a[3]);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t90, t49);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t111, a[2]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, t15, t22);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[3], r[3], tmp);
   mpz_sub (tmp, t35, t42);
   mpz_mul (tmp, t17, tmp);
   mpz_add (r[3], r[3], tmp);
   mpz_mod (r[3], r[3], q);

   mpz_mul (t174, b[3], b[3]);
   mpz_mul (t177, a[3], a[3]);
   mpz_mul (t200, t17, t17);
   /* r[4] = -9*t45*t143-9*t71*t62+9*t54*a[3]-18*t79*t62+9*t82*t174+9*t57*t177
      +9*t140*t76-6*t111*a[3]+6*t15*t62+3*t35*a[5]*b[4]+6*t90*t117-6*t114*b[3]
      -3*t26*t76+3*t45*t32-3*t53*a[6]*a[5]+t200 */
   mpz_mul (tmp, t45, t143);
   mpz_mul_si (r[4], tmp, -9);
   mpz_mul (tmp, t71, t62);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_sub (r[4], r[4], tmp);
   mpz_mul (tmp, t54, a[3]);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t79, t62);
   mpz_mul_ui (tmp, tmp, 18);
   mpz_sub (r[4], r[4], tmp);
   mpz_mul (tmp, t82, t174);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t57, t177);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t140, t76);
   mpz_mul_ui (tmp, tmp, 9);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t111, a[3]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_sub (r[4], r[4], tmp);
   mpz_mul (tmp, t15, t62);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t35, a[5]);
   mpz_mul (tmp, tmp, b[4]);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t90, t117);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t114, b[3]);
   mpz_mul_ui (tmp, tmp, 6);
   mpz_sub (r[4], r[4], tmp);
   mpz_mul (tmp, t26, t76);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[4], r[4], tmp);
   mpz_mul (tmp, t45, t32);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t53, a[6]);
   mpz_mul (tmp, tmp, a[5]);
   mpz_mul_ui (tmp, tmp, 3);
   mpz_sub (r[4], r[4], tmp);
   mpz_set (tmp, t200);
   mpz_add (r[4], r[4], tmp);
   mpz_mod (r[4], r[4], q);
#undef t3
#undef t6
#undef t7
#undef t8
#undef t11
#undef t15
#undef t16
#undef t17
#undef t20
#undef t22
#undef t26
#undef t28
#undef t31
#undef t32
#undef t35
#undef t42
#undef t45
#undef t49
#undef t53
#undef t54
#undef t57
#undef t58
#undef t62
#undef t71
#undef t76
#undef t79
#undef t82
#undef t83
#undef t86
#undef t89
#undef t90
#undef t93
#undef t96
#undef t111
#undef t114
#undef t117
#undef t121
#undef t140
#undef t143
#undef t174
#undef t177
#undef t200

  for (i=0;i<Ntmp;i++)
    mpz_clear (t[i]);
#undef Ntmp
}
#endif

/* this code was obtained as follows:
   P := a0+a1*t+a2*t^2:
   A := a[1]+a[2]*t+a[3]*t^2+a[6]*(P+3*u)+a[7]*t*(P+3*u)+a[9]*(P+3*u)^2:
   B := b[1]+b[2]*t+b[3]*t^2+b[6]*(P+3*u)+b[7]*t*(P+3*u)+b[9]*(P+3*u)^2:
   r := collect(resultant(A,B,u)/81,t):
   for i from 4 by -1 to 0 do lprint(coeff(r,t,i)) od:
   Uses 210 multiplications.
 */
static void
Res22 (mpz_t *r, mpz_t *a, mpz_t *b, mpz_t q, mpz_t tmp)
{
#define Ntmp 10
  mpz_t t[10];
  int i;
  for (i=0;i<Ntmp;i++)
    mpz_init (t[i]);
  /* r4 := a[9]^2*b[3]^2+b[9]^2*a[3]^2+b[3]*b[9]*a[7]^2+b[7]^2*a[9]*a[3]
     -b[7]*b[9]*a[7]*a[3]-b[7]*a[7]*a[9]*b[3]-2*b[9]*a[3]*a[9]*b[3] mod q */
  mpz_mul (t[0], a[9], b[3]);
  mpz_mod (t[0], t[0], q);

  mpz_mul (t[1], b[9], a[3]);
  mpz_mod (t[1], t[1], q);

  mpz_mul (r[4], t[0], t[0]);
  mpz_addmul (r[4], t[1], t[1]);
  mpz_mul (tmp, b[3], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, b[7], b[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, a[3]);
   mpz_add (r[4], r[4], tmp);
   mpz_mul (tmp, t[1], b[7]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_sub (r[4], r[4], tmp);
   mpz_mul (tmp, b[7], a[7]);
   mpz_submul (r[4], tmp, t[0]);
   mpz_mul (tmp, t[1], t[0]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[4], r[4], tmp);
   mpz_mod (r[4], r[4], q);
   /* 22 muls */

   /* r3 := b[2]*b[9]*a[7]^2+b[7]^2*a[9]*a[2]+2*b[9]^2*a[2]*a[3]
        +2*a[9]^2*b[2]*b[3]-b[7]*a[7]*a[9]*b[2]-b[6]*a[7]*a[9]*b[3]
        -b[7]*a[6]*a[9]*b[3]-b[6]*b[9]*a[7]*a[3]-b[7]*b[9]*a[7]*a[2]
        -b[7]*b[9]*a[6]*a[3]+2*b[3]*b[9]*a[6]*a[7]-2*b[9]*a[2]*a[9]*b[3]
        +2*b[6]*a[9]*b[7]*a[3]-2*b[9]*a[9]*b[2]*a[3] mod q */
   mpz_mul (tmp, b[2], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (r[3], tmp, a[7]);
   mpz_mul (tmp, b[7], b[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t[1], b[9]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t[0], a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, b[7], a[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[6], a[7]);
   mpz_mul (tmp, tmp, t[0]);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[7], a[6]);
   mpz_mul (tmp, tmp, t[0]);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[6], t[1]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[7], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[7], t[1]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[3], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, b[9], a[2]);
   mpz_mul (tmp, tmp, t[0]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[3], r[3], tmp);
   mpz_mul (tmp, b[6], a[9]);
   mpz_mul (tmp, tmp, b[7]);
   mpz_mul (tmp, tmp, a[3]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[3], r[3], tmp);
   mpz_mul (tmp, t[1], a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[3], r[3], tmp);
   mpz_mod (r[3], r[3], q);
   /* 48 muls */

   /* r2 := b[7]^2*a[9]*a[1]+b[1]*b[9]*a[7]^2+b[3]*b[9]*a[6]^2+a[9]^2*b[2]^2
      +b[9]^2*a[2]^2+b[6]^2*a[9]*a[3]+2*a[9]^2*b[1]*b[3]+2*b[9]^2*a[1]*a[3]
        -b[6]*b[9]*a[7]*a[2]-b[6]*a[6]*a[9]*b[3]-b[6]*a[7]*a[9]*b[2]
        -b[7]*a[6]*a[9]*b[2]-b[6]*b[9]*a[6]*a[3]-b[7]*b[9]*a[6]*a[2]
        -b[7]*a[7]*a[9]*b[1]-b[7]*b[9]*a[7]*a[1]-2*b[9]*a[1]*a[9]*b[3]
        +2*b[6]*a[9]*b[7]*a[2]-2*b[9]*a[2]*a[9]*b[2]+2*b[2]*b[9]*a[6]*a[7]
        -2*b[9]*a[9]*b[1]*a[3] mod q */
   mpz_mul (tmp, b[7], b[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (r[2], tmp, a[1]);
   mpz_mul (tmp, b[1], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, b[3], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, a[9], a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, b[9], b[9]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, b[6], b[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, a[3]);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, t[0], a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, b[9], t[1]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, b[6], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[6], a[6]);
   mpz_mul (tmp, tmp, t[0]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[6], a[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[7], a[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[6], t[1]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[7], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[7], a[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[7], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[9], a[1]);
   mpz_mul (tmp, tmp, t[0]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[6], a[9]);
   mpz_mul (tmp, tmp, b[7]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, b[9], a[2]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[2], r[2], tmp);
   mpz_mul (tmp, b[2], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[2], r[2], tmp);
   mpz_mul (tmp, t[1], a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[2], r[2], tmp);
   mpz_mod (r[2], r[2], q);
   /* 70 muls */

   /* r1 := b[6]^2*a[9]*a[2]+b[2]*b[9]*a[6]^2-b[7]*b[9]*a[6]*a[1]
        +2*a[9]^2*b[1]*b[2]-b[6]*a[6]*a[9]*b[2]-b[6]*a[7]*a[9]*b[1]
        +2*b[9]^2*a[1]*a[2]-b[7]*a[6]*a[9]*b[1]-b[6]*b[9]*a[6]*a[2]
        -b[6]*b[9]*a[7]*a[1]-2*b[9]*a[2]*a[9]*b[1]+2*b[1]*b[9]*a[6]*a[7]
        -2*b[9]*a[1]*a[9]*b[2]+2*b[6]*a[9]*b[7]*a[1] mod q */
   mpz_mul (tmp, b[6], b[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (r[1], tmp, a[2]);
   mpz_mul (tmp, b[2], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_add (r[1], r[1], tmp);
   mpz_mul (tmp, b[7], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, a[9], a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[1], r[1], tmp);
   mpz_mul (tmp, b[6], a[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[6], a[7]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[9], b[9]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[1], r[1], tmp);
   mpz_mul (tmp, b[7], a[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[6], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[2]);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[6], b[9]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[9], a[2]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[1], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[7]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[1], r[1], tmp);
   mpz_mul (tmp, b[9], a[1]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[2]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[1], r[1], tmp);
   mpz_mul (tmp, b[6], a[9]);
   mpz_mul (tmp, tmp, b[7]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_add (r[1], r[1], tmp);
   mpz_mod (r[1], r[1], q);
   /* 48 muls */

   /* r0 := b[1]*b[9]*a[6]^2+a[9]^2*b[1]^2+b[9]^2*a[1]^2-b[6]*b[9]*a[6]*a[1]
      -b[6]*a[6]*a[9]*b[1]+b[6]^2*a[9]*a[1]-2*b[9]*a[1]*a[9]*b[1] mod q */
   mpz_mul (tmp, b[1], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (r[0], tmp, a[6]);
   mpz_mul (tmp, a[9], a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_add (r[0], r[0], tmp);
   mpz_mul (tmp, b[9], b[9]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_add (r[0], r[0], tmp);
   mpz_mul (tmp, b[6], b[9]);
   mpz_mul (tmp, tmp, a[6]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_sub (r[0], r[0], tmp);
   mpz_mul (tmp, b[6], a[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_sub (r[0], r[0], tmp);
   mpz_mul (tmp, b[6], b[6]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, a[1]);
   mpz_add (r[0], r[0], tmp);
   mpz_mul (tmp, b[9], a[1]);
   mpz_mul (tmp, tmp, a[9]);
   mpz_mul (tmp, tmp, b[1]);
   mpz_mul_ui (tmp, tmp, 2);
   mpz_sub (r[0], r[0], tmp);
   mpz_mod (r[0], r[0], q);
   /* 22 muls */
  for (i=0;i<Ntmp;i++)
    mpz_clear (t[i]);
#undef Ntmp
}

/* z <-- f */
void
wclr_mpz_set_fr (mpz_t z, mpfr_t f)
{
  mp_exp_t e;

  e = mpfr_get_z_exp (z, f);
  if (e > 0)
    mpz_mul_2exp (z, z, e);
  else if (e < 0)
    mpz_div_2exp (z, z, -e);
}

mpz_t*
init_vec (unsigned long n)
{
  mpz_t *a;
  unsigned long i;

  a = (mpz_t*) MALLOC (n * sizeof (mpz_t));
  if (a == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  for (i=0; i<n; i++)
    mpz_init (a[i]);
  return a;
}

void
clear_vec (mpz_t *a, unsigned long n)
{
  unsigned long i;

  for (i=0; i<n; i++)
    mpz_clear (a[i]);
  FREE (a);
}

void
set_vec (mpz_t *a, mpz_t *b, unsigned long n)
{
  unsigned long i;

  for (i=0; i<n; i++)
    mpz_set (a[i], b[i]);
}

void
init_matrix (mat_ZZ *B, unsigned long n)
{
  unsigned long i;

  B->coeff = (mpz_t**) MALLOC ((n + 1) * sizeof (mpz_t*));
  if (B->coeff == NULL)
    {
      fprintf (stderr, "malloc failed\n");
      exit (1);
    }
  for (i=0; i<=n; i++)
    B->coeff[i] = init_vec (n + 1);
}

void
clear_matrix (mat_ZZ *B, unsigned long n)
{
  unsigned long i;

  for (i=0; i<=n; i++)
    clear_vec (B->coeff[i], n + 1);
  FREE (B->coeff);
}

static void
set_matrix (mat_ZZ *B, unsigned long t, mpz_t C, mpz_t a, mpz_t b, mpz_t c)
{
  mpz_t **L = B->coeff;

#define R0 5 /* t^0 */
#define R1 0 /* t^1 */
#define R2 1 /* t^2 */
#define R3 3 /* t^3 */
#define R4 2 /* t^4 */
#define R5 4 /* P+u */
#define C0 5 /* 1 */
#define C1 1 /* t */
#define C2 2 /* t^2 */
#define C3 0 /* t^3 */
#define C4 4 /* t^4 */
#define C5 3 /* u */
  mpz_set (L[R1][C1], C);                /* C */
  mpz_mul_2exp (L[R2][C2], L[R1][C1], t);   /* T*C */
  mpz_set (L[R3][C1], b);                 /* b */
  mpz_set (L[R3][C2], c);                 /* c */
  mpz_set_ui (L[R3][C5], 3);             /* 3 */
  mpz_set (L[R4][C1], a);                /* a */
  mpz_set (L[R4][C2], L[R3][C1]);        /* b */
  mpz_set (L[R4][C3], L[R3][C2]);        /* c */
  mpz_mul (L[R5][C1], a, b);              /* a*b */
  mpz_mul_2exp (L[R5][C1], L[R5][C1], 1);   /* 2*a*b */
  mpz_mul (L[R5][C2], a, c);              /* a*c */
  mpz_mul_2exp (L[R5][C2], L[R5][C2], 1);   /* 2*a*c */
  mpz_addmul (L[R5][C2], b, b);           /* 2*a*c+b^2 */
  mpz_mul (L[R5][C3], b, c);             /* b*c */
  mpz_mul_2exp (L[R5][C3], L[R5][C3], 1); /* 2*b*c */
  mpz_mul (L[R5][C4], c, c);             /* c^2 */
  mpz_mul_ui (L[R5][C5], a, 6);          /* 6*a */
#undef R0
#undef R1
#undef R2
#undef R3
#undef R4
#undef R5
#undef C0
#undef C1
#undef C2
#undef C3
#undef C4
#undef C5

  B->NumRows = ROWS;
  B->NumCols = 9;
}

/* rows 0..n-1 are multiplied by C */
void
lift_matrix (mpz_t **B, unsigned int n, mpz_t C)
{
  unsigned int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      mpz_mul (B[i][j], B[i][j], C);
}

/* rows 0..n-1 are multiplied by 2^t */
static void
lift_matrix_2exp (mpz_t **B, unsigned int n, unsigned int t)
{
  unsigned int i, j;

  for (i = 0; i < n; i++)
    for (j = 0; j < n; j++)
      mpz_mul_2exp (B[i][j], B[i][j], t);
}

/* column 5->0 0->3, 3->5 */
static void
permute_back_columns (mpz_t **B, mpz_t a, mpz_t c, mpz_t C2)
{
  unsigned int i;

  for (i = 0; i < ROWS; i++)
    {
      /* coefficient of 1 is a*([u]-3*a*[t^4]/c^2)/3 */
      mpz_divexact (B[i][5], B[i][4], c);
      mpz_divexact (B[i][5], B[i][5], c);
      mpz_mul (B[i][5], B[i][5], a);
      mpz_mul_ui (B[i][5], B[i][5], 3);
      mpz_sub (B[i][5], B[i][3], B[i][5]);
      mpz_divexact_ui (B[i][5], B[i][5], 3);
      mpz_mul (B[i][5], B[i][5], a);
      /* reduce coeff. of 1 mod C2 */
      mpz_fdiv_r (B[i][5], B[i][5], C2);
      mpz_sub (B[i][6], B[i][5], C2);
      if (mpz_cmpabs (B[i][6], B[i][5]) < 0)
        mpz_set (B[i][5], B[i][6]);
      mpz_swap (B[i][0], B[i][5]);
      mpz_swap (B[i][3], B[i][5]);
    }
}

/* tmp <- abs(V[1])+...+abs(V[9]) */
static void
Norm1 (mpz_t tmp, mpz_t *V)
{
  unsigned long i;

  mpz_set_ui (tmp, 0);

  for (i = 1; i <= 9; i++)
    {
      if (mpz_cmp_ui (V[i], 0) >= 0)
        mpz_add (tmp, tmp, V[i]);
      else
        mpz_sub (tmp, tmp, V[i]);
    }
}

/* returns sign(q(a)), where n = deg(q), using auxiliary v */
static int
sign_poly (mpz_t *q, unsigned int n, mpz_t a, mpz_t v)
{
  int i;

  mpz_set (v, q[n]);
  for (i = n - 1; i >= 0; i--)
    {
      mpz_mul (v, v, a);
      mpz_add (v, v, q[i]);
    }
  return mpz_sgn (v);
}

#ifdef DEBUG
static void
Print_vec9 (mpz_t *x)
{
  mpz_out_str (stdout, 10, x[1]);
  printf ("+t*("); mpz_out_str (stdout, 10, x[2]);
  printf (")+t^2*("); mpz_out_str (stdout, 10, x[3]);
  printf (")+t^3*("); mpz_out_str (stdout, 10, x[4]);
  printf (")+t^4*("); mpz_out_str (stdout, 10, x[5]);
  printf (")+u*("); mpz_out_str (stdout, 10, x[6]);
  printf (")+t*u*("); mpz_out_str (stdout, 10, x[7]);
  printf (")+t^2*u*("); mpz_out_str (stdout, 10, x[8]);
  printf (")+u^2*("); mpz_out_str (stdout, 10, x[9]);
  printf ("):\n");
}
#endif

static void
Norm2 (mpz_t a, mpz_t *V, mpz_t C2, unsigned int t)
{
  mpz_t q[5], b, tmp;
  int i;
  unsigned int k;

  for (i = 0; i <= 4; i++)
    mpz_init (q[i]);
  mpz_init (b);
  mpz_init (tmp);

  mpz_abs (q[4], V[5]);
  mpz_abs (q[3], V[4]);
  mpz_mul_2exp (q[3], q[3], t);
  mpz_abs (q[2], V[3]);
  if (mpz_sgn (V[8]) > 0)
    mpz_add (q[2], q[2], V[8]);
  else
    mpz_sub (q[2], q[2], V[8]);
  mpz_mul_2exp (q[2], q[2], 2 * t);
  mpz_abs (q[1], V[2]);
  if (mpz_sgn (V[7]) > 0)
    mpz_add (q[1], q[1], V[7]);
  else
    mpz_sub (q[1], q[1], V[7]);
  mpz_mul_2exp (q[1], q[1], 3 * t);
  mpz_abs (q[0], V[1]);
  if (mpz_sgn (V[6]) > 0)
    mpz_add (q[0], q[0], V[6]);
  else
    mpz_sub (q[0], q[0], V[6]);
  if (mpz_sgn (V[9]) > 0)
    mpz_add (q[0], q[0], V[9]);
  else
    mpz_sub (q[0], q[0], V[9]);
  mpz_sub (q[0], q[0], C2);
  mpz_mul_2exp (q[0], q[0], 4 * t);

  mpz_set_ui (a, 1);
  mpz_mul_2exp (a, a, t); /* we assume q(2^t) <= 0 */
  mpz_mul_2exp (b, a, 1);
  k = t; /* invariant: a=2^k, b=2^(k+1) */
  while (sign_poly (q, 4, b, tmp) <= 0)
    {
      mpz_set (a, b);
      mpz_mul_2exp (b, b, 1);
      k ++;
    }

  /* now q(a) <= 0 < q(b), with a=2^k, b=2^(k+1) */
  while (k-- > t/2) /* b-a = 2^k, no need to refine too much */
    {
      mpz_add (b, a, b);
      mpz_div_2exp (b, b, 1);
      if (sign_poly (q, 4, b, tmp) <= 0) /* (a,b) -> (b,2b-a) */
        {
          mpz_submul_ui (a, b, 2); /* a-2b */
          mpz_neg (a, a);
          mpz_swap (a, b);
        }
    }

  for (i = 0; i <= 4; i++)
    mpz_clear (q[i]);
  mpz_clear (b);
  mpz_clear (tmp);
}

/* from the coefficients V[1..9] in terms of
   1, t, t^2, t^3, t^4, u, t*u, t^2*u, u^2, get the multipliers of
   T*t, (T*t)^2, (T*t)^3, (T*t)^4, P'+3u, (T*t)*(P'+3u), (T*t)^2*(P'+3u),
   (P'+3u)^2 */
static void
get_lin_coeffs (mpz_t *V, mpz_t a, mpz_t b, mpz_t c, unsigned long t, mpz_t C,
                mpz_t C2, mpz_t q, mpz_t tmp)
{
  /* a9 := v[5]/c^2, coeff of (P'+3u)^2 */
  mpz_mul (tmp, c, c);
  mpz_divexact (V[Rt0u2], V[5], tmp);

  /* a7 := (v[4]-2*b*c*a9)/(T*c*C), coeff of (T*t)*(P'+3u) */
  mpz_mul (tmp, b, c);
  mpz_mul (tmp, tmp, V[Rt0u2]);
  mpz_mul_2exp (tmp, tmp, 1);
  mpz_sub (V[Rt1u1], V[4], tmp);
  mpz_div_2exp (V[Rt1u1], V[Rt1u1], t);
  mpz_divexact (V[Rt1u1], V[Rt1u1], c);
  mpz_divexact (V[Rt1u1], V[Rt1u1], C);

  /* a6 := (v[6]-6*a*a9)/(3*C), coeff of (P'+3u) */
  mpz_mul (tmp, a, V[Rt0u2]);
  mpz_mul_ui (tmp, tmp, 6);
  mpz_sub (V[Rt0u1], V[6], tmp);
  mpz_mul_ui (tmp, C, 3);
  mpz_divexact (V[Rt0u1], V[Rt0u1], tmp);

  /* a3 := (v[3]-(2*a*c+b^2)*a9-T*b*C*a7-c*C*a6)/(T2*C2), coeff of (Tt)^2 */
  mpz_mul (tmp, a, c);
  mpz_mul_2exp (tmp, tmp, 1);
  mpz_mul (V[8], b, b);
  mpz_add (tmp, tmp, V[8]);
  mpz_mul (tmp, tmp, V[Rt0u2]);
  mpz_sub (V[Rt2u0], V[3], tmp);
  mpz_mul (tmp, b, C);
  mpz_mul (tmp, tmp, V[Rt1u1]);
  mpz_mul_2exp (tmp, tmp, t);
  mpz_sub (V[Rt2u0], V[Rt2u0], tmp);
  mpz_mul (tmp, c, C);
  mpz_mul (tmp, tmp, V[Rt0u1]);
  mpz_sub (V[Rt2u0], V[Rt2u0], tmp);
  mpz_div_2exp (V[Rt2u0], V[Rt2u0], 2 * t);
  mpz_divexact (V[Rt2u0], V[Rt2u0], C2);

  /* a2 := (v[2]-2*a*b*a9-T*a*C*a7-b*C*a6)/(T*C2), coeff of (Tt) */
  mpz_mul (tmp, a, b);
  mpz_mul (tmp, tmp, V[Rt0u2]);
  mpz_mul_2exp (tmp, tmp, 1);
  mpz_sub (V[Rt1u0], V[2], tmp);
  mpz_mul (tmp, a, C);
  mpz_mul (tmp, tmp, V[Rt1u1]);
  mpz_mul_2exp (tmp, tmp, t);
  mpz_sub (V[Rt1u0], V[Rt1u0], tmp);
  mpz_mul (tmp, b, C);
  mpz_mul (tmp, tmp, V[Rt0u1]);
  mpz_sub (V[Rt1u0], V[Rt1u0], tmp);
  mpz_div_2exp (V[Rt1u0], V[Rt1u0], t);
  mpz_divexact (V[Rt1u0], V[Rt1u0], C2);

  /* a1 := (v[1]-a^2*a9-a*C*a6)/C2, coeff of 1 */
  mpz_mul (tmp, a, a);
  mpz_mul (tmp, tmp, V[Rt0u2]);
  mpz_sub (V[Rt0u0], V[1], tmp);
  mpz_mul (tmp, a, C);
  mpz_mul (tmp, tmp, V[Rt0u1]);
  mpz_sub (V[Rt0u0], V[Rt0u0], tmp);
  mpz_divexact (V[Rt0u0], V[Rt0u0], C2);

  mpz_mod (V[Rt0u0], V[Rt0u0], q);
  mpz_mod (V[Rt1u0], V[Rt1u0], q);
  mpz_mod (V[Rt2u0], V[Rt2u0], q);
  mpz_mod (V[Rt0u1], V[Rt0u1], q);
  mpz_mod (V[Rt1u1], V[Rt1u1], q);
  mpz_mod (V[Rt0u2], V[Rt0u2], q);
}

void
Mod (mpz_t *X, int d, mpz_t q)
{
  int i;
  for (i = 0; i <= d; i++)
    mpz_mod (X[i], X[i], q);
}

/* make Res monic, where deg(Res)=d */
void
MakeMonic (mpz_t *Res, int d, mpz_t q)
{
  int i;

  if (mpz_invert (Res[d], Res[d], q) == 0)
    abort ();
  for (i = 0; i < d; i++)
    mpz_mul (Res[i], Res[i], Res[d]);
  Mod (Res, d - 1, q);
  mpz_set_ui (Res[d], 1);
}

/* X <- X - X[dX]*t^(dX-d)*Res, return degree of remainder */
int
Reduce (mpz_t *X, unsigned long dX, mpz_t *Res, int d, mpz_t q)
{
  int i;
  mpz_t tmp;

  mpz_init (tmp);
  mpz_mod (X[dX], X[dX], q);
  for (i = 0; i < d; i++)
    {
      mpz_mul (tmp, X[dX], Res[i]);
      mpz_sub (X[dX - d + i], X[dX - d + i], tmp);
    }
  /*  mpz_set_ui (X[dX], 0); */ /* don't zero to get the quotient in place */
  for (i = dX - 1; i >= 0; i--)
    {
      mpz_mod (X[i], X[i], q);
      if (mpz_cmp_ui (X[i], 0) != 0)
        break;
    }
  mpz_clear (tmp);
  return i;
}

/* a <- GCD(a, b) where deg(a)=da > deg(b)=db.
   Return degree of GCD.
*/
unsigned long
GCD (mpz_t *a, int da, mpz_t *b, int db, mpz_t q)
{
  int i;

  while (db >= 0)
    {
      if (mpz_cmp_ui (b[db], 1) != 0)
        MakeMonic (b, db, q);
      while (da >= db)
        da = Reduce (a, da, b, db, q);
      /* now da < db */
      /* swap polynomials to ensure result is in a */
      for (i = 0; i <= db; i++)
        mpz_swap (a[i], b[i]);
      i = da;
      da = db;
      db = i;
    }
  return da;
}

/* X <- (t+a)^e mod (Res,q), return result degree */
int
Powmod (mpz_t *X, int a, mpz_t e, mpz_t *Res, int d, mpz_t q)
{
  mpz_t tmp;
  int l = mpz_sizeinbase (e, 2) - 1, i, dX;

  mpz_set_ui (X[0], a);
  mpz_set_ui (X[1], 1); /* X = x+a */
  dX = 1;
  mpz_init (tmp);
  while (--l >= 0)
    {
      /* square X */
      if (dX >= 3)
        {
          mpz_mul (X[6], X[3], X[3]);
          mpz_mul (X[5], X[2], X[3]);
          mpz_mul_2exp (X[5], X[5], 1);
        }
      if (dX >= 2)
        mpz_mul (X[4], X[2], X[2]);
      if (dX >= 3)
        {
          mpz_mul (tmp, X[1], X[3]);
          mpz_mul_2exp (tmp, tmp, 1);
          mpz_add (X[4], X[4], tmp);
        }
      if (dX >= 3)
        mpz_mul (X[3], X[0], X[3]);
      else
        mpz_set_ui (X[3], 0);
      if (dX >= 2)
        {
          mpz_mul (tmp, X[1], X[2]);
          mpz_add (X[3], X[3], tmp);
          mpz_mul_2exp (X[3], X[3], 1);
          mpz_mul (X[2], X[0], X[2]);
          mpz_mul_2exp (X[2], X[2], 1);
        }
      else
        mpz_set_ui (X[2], 0);
      if (dX >= 1)
        {
          mpz_mul (tmp, X[1], X[1]);
          mpz_add (X[2], X[2], tmp);
          mpz_mul (X[1], X[0], X[1]);
          mpz_mul_2exp (X[1], X[1], 1);
        }
      mpz_mul (X[0], X[0], X[0]);
      dX = 2 * dX;
      while (dX >= d)
        dX = Reduce (X, dX, Res, d, q);
      Mod (X, dX, q);
      /* multiply by t+a */
      if (mpz_tstbit (e, l))
        {
          mpz_set_ui (X[dX + 1], 0);
          for (i = dX; i >= 0; i--)
            {
              mpz_mul_ui (X[i + 1], X[i + 1], a);
              mpz_add (X[i + 1], X[i + 1], X[i]);
            }
          mpz_mul_ui (X[0], X[0], a);
          dX ++;
          if (dX >= d)
            dX = Reduce (X, dX, Res, d, q);
          Mod (X, dX, q);
        }
    }
  mpz_clear (tmp);

  return Degree (X, dX);
}

/* Res <- gcd(Res, x^q-x), where Res is monic of degree d, 2 <= d <= 4.
   Return degree of gcd */
unsigned long
DDF (mpz_t *Res, unsigned long d, mpz_t q)
{
  mpz_t *X;
  unsigned long dg;
  int dX;

  X = init_vec (2 * d);
  dX = Powmod (X, 0, q, Res, d, q);
  /* subtract x */
  if (dX >= 1)
    mpz_sub_ui (X[1], X[1], 1);
  else
    {
      mpz_set_si (X[1], -1);
      dX = 1;
    }
  /* compute gcd */
  Mod (X, dX, q);
  dX = Degree (X, dX);
  dg = GCD (Res, d, X, dX, q);
#ifdef VERBOSE
  printf ("gcd="); Print_vec (Res, dg + 1);
#endif
  clear_vec (X, 2 * d);

  return dg;
}

/* Res[i] <- Res[i] - q*round(Res[i]/q), where e = (q-1)/2 */
void
centermod (mpz_t *Res, int d, mpz_t q)
{
  mpz_t e;
  int i;

  mpz_init (e);
  mpz_sub_ui (e, q, 1);
  mpz_div_2exp (e, e, 1); /* e = (q-1)/2 */
  for (i=0; i<d; i++)
    {
      mpz_mod (Res[i], Res[i], q);
      if (mpz_cmpabs (Res[i], e) > 0)
        mpz_sub (Res[i], Res[i], q);
    }
  mpz_clear (e);
}

/* splits a of degree 2 <= da <= 4 into linear factors */
void
Split (mpz_t *a, int da, mpz_t q)
{
  mpz_t *h, *copy_a;
  int i = 0, dh, j, d0 = da;
  mpz_t e;

  mpz_init (e);
  mpz_sub_ui (e, q, 1);
  mpz_div_2exp (e, e, 1);
  h = init_vec (2 * da);
  copy_a = init_vec (da + 1);
  while (da > 1)
    {
      do
        {
          dh = Powmod (h, ++i, e, a, da, q);
          if (dh >= 0)
            mpz_sub_ui (h[0], h[0], 1);
          else
            {
              mpz_set_si (h[0], -1);
              dh = 0;
            }
          Mod (h, dh, q);
          dh = Degree (h, dh);
          if (dh > 0)
            {
              set_vec (copy_a, a, da + 1);
              dh = GCD (copy_a, da, h, dh, q);
              set_vec (h, copy_a, dh + 1);
            }
        }
      while (dh != 1);
      MakeMonic (h, dh, q);
      /* divide in place a by h */
      j = da;
      while (j >= 1)
        j = Reduce (a, j, h, 1, q);
      mpz_neg (a[0], h[0]);
      a ++;
      da --;
    }
  mpz_neg (a[0], a[0]);
  clear_vec (h, 2 * d0);
  clear_vec (copy_a, d0 + 1);
  mpz_clear (e);
}

#ifdef DEBUG
static void
Print_vec (mpz_t *x, long n)
{
  long i;

  printf("[");
  for (i=0; i<n;)
    {
      mpz_out_str (stdout, 10, x[i]);
      if (++i < n)
        printf (", ");
    }
  printf ("]\n");
}
#endif

/* put in Res[0]..Res[n-1] the roots of Res[0]+Res[1]*t+...+Res[d]*t^d mod q,
   and return the number n of roots */
unsigned int
Roots (mpz_t *Res, int d, mpz_t q)
{
  if (d == 0)
    return 0;

  MakeMonic (Res, d, q);

#ifdef VERBOSE
  printf ("Resultant="); Print_vec (Res, d + 1);
#endif

  if (d > 1)
    {
      d = DDF (Res, d, q);
#ifdef VERBOSE
      printf ("after DDF: "); Print_vec (Res, d + 1);
#endif
    }

  if (d == 1)
    mpz_neg (Res[0], Res[0]);
  else
    Split (Res, d, q);

#ifdef VERBOSE
  printf ("Roots are "); Print_vec (Res, d);
#endif

  return d;
}

/* M = 2^m, T = 2^t
   Return -1 in case of failure, otherwise number of worst cases found.
   q = nextprime(2^(EXTRA+1)*T).
   Clobbers P[].
 */
static int
slz (mpz_t *Res, mpfr_t *P, unsigned int m, unsigned int t, mpz_t q,
     double eps, unsigned long *calls, mpz_t width)
{
  double Mprime;
  unsigned long i, j;
  mpz_t C, C2, a, b, c, tmp, **V, norm2[ROWS];
  mat_ZZ B[1];
  int ok[ROWS], dRes, perm[ROWS+1];

  mpz_init (C);
  mpz_init (C2);
  mpz_init (a);
  mpz_init (b);
  mpz_init (c);
  mpz_init (tmp);
  init_matrix (B, 9);
  for (i=0; i<ROWS; i++)
    mpz_init (norm2[i]);

  /* Mprime := floor(1/2/(1/M+eps)) */
  for (Mprime = 1.0, i = 0; i < m; Mprime *= 2.0, i++); /* 2^m */
  eps = 1 / (Mprime - 1.0) + eps;
  Mprime = floor (0.5 / eps);
#ifdef VERBOSE
  printf ("eps=%1.20e M'=%1.0f\n", eps, Mprime);
#endif

  /* C:=3*Mprime */
  mpz_set_d (C, Mprime);
  i = mpz_sizeinbase (C, 2) / 2;
  mpz_tdiv_q_2exp (C, C, i);
  mpz_mul_2exp (C, C, i);
  mpz_mul_ui (C, C, 3);
#ifdef VERBOSE
  printf ("C="); mpz_out_str (stdout, 10, C); printf ("\n");
#endif
  mpz_mul (C2, C, C);

  /* a := round(C*frac(coeff(P,t,0))) */
  mpfr_frac (P[0], P[0], GMP_RNDN);
  mpfr_mul_z (P[0], P[0], C, GMP_RNDN);
  wclr_mpz_set_fr (a, P[0]);
#ifdef VERBOSE
  printf ("a="); mpz_out_str (stdout, 10, a); printf ("\n");
#endif

  /* b := round(C*coeff(P,t,1)*T) */
  mpfr_mul_z (P[1], P[1], C, GMP_RNDN);
  mpfr_mul_2exp (P[1], P[1], t, GMP_RNDN);
  mpfr_round (P[1], P[1]);
  wclr_mpz_set_fr (b, P[1]);
#ifdef VERBOSE
  printf ("b="); mpz_out_str (stdout, 10, b); printf ("\n");
#endif

  /* c := round(C*coeff(P,t,2)*T^2 */
  mpfr_mul_z (P[2], P[2], C, GMP_RNDN);
  mpfr_mul_2exp (P[2], P[2], 2 * t, GMP_RNDN);
  mpfr_round (P[2], P[2]);
  wclr_mpz_set_fr (c, P[2]);
#ifdef VERBOSE
  printf ("c="); mpz_out_str (stdout, 10, c); printf ("\n");
#endif

  set_matrix (B, t, C, a, b, c);
  /* column 0:tau^3 1:tau 2:tau^2 3:ups 4:1 5:tau^4 */

  /* printf ("input matrix:\n");
     Print_maple (B->coeff, 6, 6); */
  *calls = 0;
  lll_time -= cputime ();
#ifdef CUT
  lift_matrix_2exp (B->coeff, 3, t);
  lift_matrix (B->coeff, 4, C);
  *calls += lllint2 (B->coeff, 5, 5, CUT);
#else
  *calls += lllint (B->coeff, 3, 3);
  lift_matrix_2exp (B->coeff, 3, t);
  *calls += lllint (B->coeff, 4, 4);
  lift_matrix (B->coeff, 4, C);
  *calls += lllint (B->coeff, 5, 5);
#endif
  lll_time += cputime ();
  /* column 0:tau^3 1:tau 2:tau^2 3:u 4:tau^4 5:1 */
  permute_back_columns (B->coeff, a, c, C2);
  /* column 0:1 1:tau 2:tau^2 3:tau^3 4:tau^4 5:u */
  /* printf ("output matrix:\n");
     Print_maple (B->coeff, 6, 6); */
  post_copy (B->coeff, 6, 6); /* B[0..n-1] -> B[1..n] */
  /* column 1:1 2:tau 3:tau^2 4:tau^3 5:tau^4 6:ups */
  V = B->coeff;
  /* restore 3 last columns from previous ones:
     The coefficient of u*t is 3/c times that of t^3.
     That of u*t^2 is 6/c that of t^4.
     That of u^2 is 9/c^2 that of t^4.
   */
  for (i = 1; i < ROWS; i++)
    {
      /* V[i][4] = k*(2*b*c) -> V[i][7] = k*(6*b) */
      mpz_divexact (V[i][7], V[i][4], c);
      mpz_mul_ui (V[i][7], V[i][7], 3);
      /* column 7: tau*ups */
      /* V[i][5] = k*(c^2) -> V[i][8] = k*(6*c), V[i][9] = k*(9) */
      mpz_divexact (V[i][8], V[i][5], c);
      mpz_mul_ui (V[i][8], V[i][8], 6);
      /* column 8: tau^2*ups */
      mpz_divexact (V[i][9], V[i][8], c);
      mpz_div_2exp (V[i][9], V[i][9], 1);
      mpz_mul_ui (V[i][9], V[i][9], 3);
      /* column 9: ups^2 */
    }

  for (i = 1; i < ROWS; i++)
    {
      Norm1 (tmp, V[i]);
      if (mpz_cmp (tmp, C2) < 0)
        {
          ok[i] = 1;
          Norm2 (norm2[i-1], V[i], C2, t);
        }
      else
        ok[i] = 0; /* vector is too large */
    }

  /* indirect sort: norm2[perm[i]-1] is decreasing with i */
  for (i = 1; i < ROWS; i++)
    {
      if (ok[i])
        {
          for (j = i - 1; j >= 1 && (ok[j] == 0 ||
                 (mpz_cmp (norm2[perm[j]-1], norm2[i-1]) < 0)); j--)
            perm[j+1] = perm[j];
          perm[j+1] = i;
        }
      else
        perm[i] = i;
    }

  for (dRes = -1, i = 2; i < ROWS && dRes == -1; i++)
    {
      if (ok[perm[i]])
        {
          if (ok[perm[i]] == 1)
            {
              get_lin_coeffs (V[perm[i]], a, b, c, t, C, C2, q, tmp);
              ok[perm[i]] = 2;
            }
          for (j = 1; j < i && dRes == -1; j++)
            {
              if (ok[perm[j]])
                {
                  if (ok[perm[j]] == 1)
                    {
                      get_lin_coeffs (V[perm[j]], a, b, c, t, C, C2, q, tmp);
                      ok[perm[j]] = 2;
                    }
                  Res22 (Res, V[perm[i]], V[perm[j]], q, tmp);
                  dRes = Degree (Res, 4);
                  if (dRes >= 0)
                    {
                      if (mpz_cmp (norm2[perm[i]-1], norm2[perm[j]-1]) > 0)
                        mpz_set (width, norm2[perm[j]-1]);
                      else
                        mpz_set (width, norm2[perm[i]-1]);
                      /* check width <= 2^EXTRA*T = 2^(t+EXTRA) */
                      if (mpz_sizeinbase (width, 2) > t + EXTRA)
                        {
                          mpz_set_ui (width, 1);
                          mpz_mul_2exp (width, width, t + EXTRA);
                        }
                    }
                }
            }
        }
    }

#ifdef VERBOSE
  if (dRes == -1)
    {
      fprintf (stderr, "Error, no valid vector found\n");
    }
#endif

  if (dRes > 0)
    {
      dRes = Roots (Res, dRes, q);
      centermod (Res, dRes, q);
#ifdef VERBOSE
      printf ("centered Roots are "); Print_vec (Res, dRes);
#endif
    }

  mpz_clear (C);
  mpz_clear (C2);
  mpz_clear (a);
  mpz_clear (b);
  mpz_clear (c);
  mpz_clear (tmp);
  clear_matrix (B, 9);
  for (i=0; i<ROWS; i++)
    mpz_clear (norm2[i]);

  return dRes;
}

int
Check (mpz_t *Res, int d, mpz_t t0, unsigned long int n, long int m)
{
  int i, k = 0;
  mpfr_t x, y;

  mpfr_init (x);
  mpfr_init (y);
  for (i=0; i<d; i++)
    {
      mpz_add (Res[i], Res[i], t0);
      mpfr_set_z (x, Res[i], GMP_RNDN);
      mpfr_div_2exp (x, x, n - 1, GMP_RNDN);
      mpfr_sub_ui (x, x, 1, GMP_RNDN);
      mpfr_div_2exp (x, x, 1, GMP_RNDN);
      mpfr_exp2 (x, x, GMP_RNDN);
      mpfr_mul_2exp (x, x, n + 1, GMP_RNDN);
      mpfr_round (y, x);
      mpfr_sub (y, x, y, GMP_RNDN);
      /* check if x < 1/2^m, i.e. EXP(x) <= -m */
      if (mpfr_cmp_ui (y, 0) == 0 || mpfr_get_exp (y) <= -m)
        {
          k ++;
          printf ("*** t0=");
          mpz_out_str (stdout, 10, Res[i]);
          printf (" m=");
          if (mpfr_cmp_ui (y, 0) == 0)
            printf ("inf");
          else
            printf ("%lu", -mpfr_get_exp (y));
          printf ("\n");
        }
    }
  mpfr_clear (x);
  mpfr_clear (y);

  return k;
}

int
main (int argc, char *argv[])
{
  mp_prec_t prec;
  unsigned long int n, m, t, initial_t;
  mpz_t t0, t1, T, q, *Res, tm, width;
  mpfr_t P[3];
  mpfr_t x;
  mpfr_t ln2;
  double eps;
  int d, verbose = 0, first = 0;
  unsigned long tries = 0; /* number of calls to slz */
  unsigned long failures = 0; /* number of failures */
  double total_T = 0.0; /* total width */
  unsigned long calls, tot_calls = 0; /* calls to reduce/swap in lllint */
  int total_time;
#ifdef WITH_MPI
  mpz_t s0, s1;
  int myid, numprocs;

  MPI_Init (&argc, &argv);
  MPI_Comm_size (MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank (MPI_COMM_WORLD, &myid);
#endif

  total_time = cputime ();
  printf ("Welcome to WCLR version %s ", VERSION);
  init_lll ();

  lll_time = 0;

  if (argc > 1 && strcmp (argv[1], "-v") == 0)
    {
      verbose = 1;
      argv ++;
      argc --;
    }

  if (argc > 1 && strcmp (argv[1], "-first") == 0)
    {
      first = 1;
      argv ++;
      argc --;
    }

  if (argc != 7)
    {
      fprintf (stderr, "Usage: wclr22 [-v] [-first] prec n m t t0 t1\n");
      exit (1);
    }

  prec = atoi (argv[1]);
  mpfr_set_default_prec (prec);

  n = atoi (argv[2]);
  m = atoi (argv[3]);
  t = atoi (argv[4]);
  initial_t = t;

  mpz_init (t0);
  mpz_init (t1);
  mpz_init (tm);
  mpz_init (T);
  mpz_init (q);
  mpz_init (width);

  /* T = 2^t */
  mpz_set_ui (T, 1);
  mpz_mul_2exp (T, T, t);

  /* q = nextprime(2*T*2^EXTRA) */
  mpz_mul_2exp (q, T, 1 + EXTRA);
  mpz_nextprime (q, q);

#ifdef VERBOSE
  printf ("q=");
  mpz_out_str (stdout, 10, q);
  printf ("\n");
#endif

  mpfr_init (ln2);
  mpfr_init (x);
  mpfr_init (P[0]);
  mpfr_init (P[1]);
  mpfr_init (P[2]);

#ifdef WITH_MPI
  mpz_init (s0);
  mpz_init (s1);
  mpz_set_str (s0, argv[5], 10);
  mpz_set_str (s1, argv[6], 10);
  mpz_mul_ui (t0, s0, numprocs - myid);
  mpz_addmul_ui (t0, s1, myid);
  mpz_sub (t1, t0, s0);
  mpz_add (t1, t1, s1);
  mpz_tdiv_q_ui (t0, t0, numprocs);
  mpz_tdiv_q_ui (t1, t1, numprocs);
  printf ("numprocs = %d, myid = %d, t0 = ", numprocs, myid);
  mpz_out_str (stdout, 10, t0);
  printf (", t1 = ");
  mpz_out_str (stdout, 10, t1);
  printf ("\n");
#else
  mpz_set_str (t0, argv[5], 10);
  mpz_set_str (t1, argv[6], 10);
#endif

  /* compute log(2)/N in ln2 */
  mpfr_const_log2 (ln2, GMP_RNDN);
  mpfr_div_2exp (ln2, ln2, n, GMP_RNDN);

  /* compute global error bound:
     3rd coeff is 1/6*2^(t0/N-1/2)*ln(2)^3/N^3 <= 1/6*2^(1/2)*ln(2)^3/N^3 */
  mpfr_set_d (x, 1.0 / 6.0 * pow (log (2), 3.0), GMP_RNDU);
  mpfr_div_2exp (x, x, 2 * n - 3 * t - 1, GMP_RNDU);
  eps = mpfr_get_d (x, GMP_RNDU);

  Res = init_vec (5);

  while (mpz_cmp (t0, t1) < 0)
    {
      if (verbose)
        {
          printf ("t0=");
          mpz_out_str (stdout, 10, t0);
          printf ("\n");
        }

      mpz_add (tm, t0, T); /* tm is middle of current interval */

      /* compute x = 2^(tm/N-1/2) = 2^((tm/(N/2)-1)/2) */
      mpfr_set_z (x, tm, GMP_RNDN);
      mpfr_div_2exp (x, x, n - 1, GMP_RNDN);
      mpfr_sub_ui (x, x, 1, GMP_RNDN);
      mpfr_div_2exp (x, x, 1, GMP_RNDN);
      mpfr_exp2 (x, x, GMP_RNDN);

      /* compute series expansion */
      mpfr_mul_2exp (P[0], x, n + 1, GMP_RNDN); /* 2*N*f(tm) */
      mpfr_mul (P[1], P[0], ln2, GMP_RNDN); /* 2*f(tm)*ln(2) */
      mpfr_mul (P[2], P[1], ln2, GMP_RNDN);
      mpfr_div_2exp (P[2], P[2], 1, GMP_RNDN); /* f(tm)*ln(2)^2/N */

      d = slz (Res, P, m, t, q, eps * mpfr_get_d (x, GMP_RNDU), &calls, width);
      tot_calls += calls;
      tries ++;
      if (d == -1)
        {
          failures ++;
          if (verbose || t == 0)
            {
              fprintf (stderr, "FAIL at t0=");
              mpz_out_str (stderr, 10, t0);
              fprintf (stderr, " with T=");
              mpz_out_str (stderr, 10, T);
              fprintf (stderr, "\n");
              if (t == 0)
                goto end;
            }
          t --;
          mpz_div_2exp (T, T, 1);
        }
      else /* success */
        {
          total_T += pow (2.0, (double) t) + mpz_get_d (width);
          /* check potential solutions */
          if (Check (Res, d, tm, n, m) && first)
            break;

          mpz_add (t0, tm, width);

          if (t < initial_t)
            {
              t ++;
              mpz_mul_2exp (T, T, 1);
            }
        }
    }

  total_time = cputime () - total_time;

  printf ("%lu calls to slz, %lu failures, average width %1.0f\n", tries,
          failures, total_T / (double) tries);
  printf ("%lu calls to reduce/swap (average %lu)\n", tot_calls,
          tot_calls / tries);

  printf ("time = %dms (lll time = %dms)\n", total_time, lll_time);

  clear_vec (Res, 5);

  mpfr_clear (P[0]);
  mpfr_clear (P[1]);
  mpfr_clear (P[2]);
  mpfr_clear (x);
  mpfr_clear (ln2);

  mpz_clear (t0);
  mpz_clear (t1);
  mpz_clear (tm);
  mpz_clear (T);
  mpz_clear (q);
  mpz_clear (width);

 end:
#ifdef WITH_MPI
  MPI_Finalize ();
#endif

  return 0;
}
